-----------------------------------------------------------------------
-- Upvalued Lua API.
-----------------------------------------------------------------------
local _G = getfenv(0)

-- Functions
local pairs = _G.pairs
local tonumber = _G.tonumber
local type = _G.type

-- Libraries
local math = _G.math
local table = _G.table

-----------------------------------------------------------------------
-- AddOn namespace.
-----------------------------------------------------------------------
local ADDON_NAME, private = ...

local LibStub = _G.LibStub

local BTC = LibStub("AceAddon-3.0"):NewAddon(ADDON_NAME, "AceEvent-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale(ADDON_NAME)

local LQT = LibStub:GetLibrary("LibQTip-1.0")
local LDBIcon = LibStub("LibDBIcon-1.0")

local ADDON_NAME_SHORT = "BTC"
_G[ADDON_NAME_SHORT] = BTC

local addon = BTC

local addonversion
do
	local debug_version = false
	local alpha_version = false

	--[===[@debug@
	debug_version = true
	--@end-debug@]===]

	--@alpha@
	alpha_version = true
	--@end-alpha@

	local addon_version = _G.GetAddOnMetadata(ADDON_NAME, "Version")
	addonversion = debug_version and "Devel" or (alpha_version and addon_version .. "-Alpha") or addon_version
end -- do-block
local addonauthor = "James D. Callahan III"
local builddate = "20121025011443"

------------------------------------------------------------------------------
-- Constants.
------------------------------------------------------------------------------
local COOLDOWN_SPELL_MAP = {
	-- Initialize my cooldowns list
	-- Alchemy
	-- Northrend Alchemy Research
	[60893] = true, -- Northrend Alchemy Research
	-- Metals (shared cooldown)
	[11479] = L["Elements"], -- Transmute: Iron to Gold
	[11480] = L["Elements"], -- Transmute: Mithril to Truesilver
	-- rm'd in 3.3.0
	-- [60350] = L["Elements"],		-- Transmute: Titanium
	-- Elements (shared cooldown)
	[17559] = L["Elements"], -- Transmute: Air to Fire
	[17566] = L["Elements"], -- Transmute: Earth to Life
	[17561] = L["Elements"], -- Transmute: Earth to Water
	--[[ No cooldown on Elemental Fire
			 [25146] = L["Elements"],
			 ]] --
	[17560] = L["Elements"], -- Transmute: Fire to Earth
	[17565] = L["Elements"], -- Transmute: Life to Earth
	[17563] = L["Elements"], -- Transmute: Undeath to Water
	[17562] = L["Elements"], -- Transmute: Water to Air
	[17564] = L["Elements"], -- Transmute: Water to Undeath
	-- Primals
	[28566] = L["Elements"], -- Transmute: Primal Air to Fire
	[28585] = L["Elements"], -- Transmute: Primal Earth to Life
	[28567] = L["Elements"], -- Transmute: Primal Earth to Water
	[28568] = L["Elements"], -- Transmute: Primal Fire to Earth
	[28583] = L["Elements"], -- Transmute: Primal Fire to Mana
	[28584] = L["Elements"], -- Transmute: Primal Life to Earth
	[28582] = L["Elements"], -- Transmute: Primal Mana to Fire
	[28580] = L["Elements"], -- Transmute: Primal Shadow to Water
	[28569] = L["Elements"], -- Transmute: Primal Water to Air
	[28581] = L["Elements"], -- Transmute: Primal Water to Shadow
	-- Eternals
	[53777] = L["Elements"], -- Transmute: Eternal Air to Earth
	[52776] = L["Elements"], -- Transmute: Eternal Air to Water
	[53781] = L["Elements"], -- Transmute: Eternal Earth to Air
	[53782] = L["Elements"], -- Transmute: Eternal Earth to Shadow
	[53775] = L["Elements"], -- Transmute: Eternal Fire to Life
	[53774] = L["Elements"], -- Transmute: Eternal Fire to Water
	[53773] = L["Elements"], -- Transmute: Eternal Life to Fire
	[53771] = L["Elements"], -- Transmute: Eternal Life to Shadow
	[54020] = L["Elements"], -- Transmute: Eternal Might
	[53779] = L["Elements"], -- Transmute: Eternal Shadow to Earth
	[52780] = L["Elements"], -- Transmute: Eternal Shadow to Life
	[53783] = L["Elements"], -- Transmute: Eternal Water to Air
	[53784] = L["Elements"], -- Transmute: Eternal Water to Fire
	-- Epic gems
	[66658] = L["Elements"], -- Transmute : Ametrine
	[66659] = L["Elements"], -- Transmute : Cardinal Ruby
	[66660] = L["Elements"], -- Transmute : King's Amber
	[66662] = L["Elements"], -- Transmute : Dreadstone
	[66663] = L["Elements"], -- Transmute : Majestic Zircon
	[66664] = L["Elements"], -- Transmute : Eye of Zul
	-- Diamonds (shared cooldown)
	--[[ It appears these were removed as well
			 [32765] = L["Diamonds"],		-- Transmute: Earthstorm Diamond
			 [32766] = L["Diamonds"],		-- Transmute: Skyfire Diamond
			 [57425] = L["Diamonds"],		-- Transmute: Skyflare Diamond
			 [57427] = L["Diamonds"],		-- Transmute: Earthsiege Diamond
			 ]] --
	-- New Cata Elemental transmutes
	[78866] = L["Elements"], -- Transmute: Living Elements
	[80243] = L["Elements"], -- Transmute: Truegold
	[80244] = L["Elements"], -- Transmute: Pyrium Bar
	-- MoP
	[114780] = L["Elements"], -- Transmute: Living Steel

	-- New Cata Gems
	--[[ Removed somewhere around 4.1
			 [80245] = L["Cataclysm Blue Gems"],		-- Transmute: Inferno Ruby
			 [80246] = L["Cataclysm Blue Gems"],		-- Transmute: Ocean Sapphire
			 [80247] = L["Cataclysm Blue Gems"],		-- Transmute: Amberjewel
			 [80248] = L["Cataclysm Blue Gems"],		-- Transmute: Demonseye
			 [80250] = L["Cataclysm Blue Gems"],		-- Transmute: Ember Topaz
			 [80251] = L["Cataclysm Blue Gems"],		-- Transmute: Dream Emerald
			 ]] --

	-- Enchanting
	[28027] = true, -- Prismatic Sphere
	[28028] = true, -- Void Sphere

	-- Jewelcrafting
	[47280] = true, -- Brilliant Glass
	[62242] = true, -- Icy Prism
	[73478] = true, -- Fire Prism

	-- MoP Jewelcrafting (shared cooldown)
	[131691] = L["Facets of Research"], -- Imperial Amethyst
	[131686] = L["Facets of Research"], -- Primordial Ruby
	[131593] = L["Facets of Research"], -- River's Heart
	[131695] = L["Facets of Research"], -- Sun's Radiance
	[131690] = L["Facets of Research"], -- Vermilion Onyx
	[131688] = L["Facets of Research"], -- Wild Jade

	-- Tailoring
	-- Wrath of the Lich King
	[56005] = true, -- Glacial Bag
	-- Cataclysm
	[75141] = true, -- Dream of Skywall
	[75145] = true, -- Dream of Ragnaros
	[75144] = true, -- Dream of Hyjal
	[75142] = true, -- Dream of Deepholm
	[75146] = true, -- Dream of Azshara
	-- Mists of Pandaria
	[125557] = true, -- Imperial Silk
	-- Inscription
	[61288] = true, -- Minor Inscription Research
	[61177] = true, -- Northrend Inscription Research
	[86654] = true, -- Horde Forged Documents
	[89244] = true, -- Alliance Forged Documents
	[112996] = true, -- Scroll of Wisdom
}


------------------------------------------------------------------------------
-- Variables.
------------------------------------------------------------------------------
local db
local tooltip

local primary = {}
local secondary = {}


------------------------------------------------------------------------------
-- Helper functions.
------------------------------------------------------------------------------
do
	local CHAT_MESSAGE_FORMAT = "\124TInterface\\Addons\\" .. ADDON_NAME .. "\\icon:" .. "%d" .. "\124t" .. ADDON_NAME_SHORT .. ": " .. "%s"

	function addon.Print(message)
		if not message then
			return
		end
		_G.DEFAULT_CHAT_FRAME:AddMessage(CHAT_MESSAGE_FORMAT:format(math.ceil(_G.select(2, _G.DEFAULT_CHAT_FRAME:GetFont())), message), 0.6, 1.0, 1.0)
	end
end -- do-block

function addon.Debug(message)
	if not message or not db.options.debug then
		return
	end
	addon.Print("DEBUG: " .. message)
end

-- coloring functions for tooltip text
local function Red(text) return "|cffff0000" .. text .. "|r" end

local function Orange(text) return "|cffdea51c" .. text .. "|r" end

local function Yellow(text) return "|cffffff00" .. text .. "|r" end

local function Green(text) return "|cff00ff00" .. text .. "|r" end

local function White(text) return "|cffffffff" .. text .. "|r" end

local function BTC_prettyTime(seconds)
	local remaining = seconds
	local output = ""

	if remaining >= 86400 then
		output = output .. math.floor(remaining / 86400) .. "d"
		remaining = remaining % 86400
	end

	if remaining >= 3600 then
		output = output .. math.floor(remaining / 3600) .. "h"
		remaining = remaining % 3600
	end

	if remaining >= 60 then
		output = output .. math.floor(remaining / 60) .. "m"
	end
	return output
end

local function PercentColorText(min, max, text)
	local num = min + 0
	local den = max + 0

	if den == 0 then
		return Green(text)
	end
	local perc = num / den
	addon.Debug(min .. "/" .. max .. " = " .. perc)

	if perc < 0.25 then
		return Red(text)
	elseif perc < 0.5 then
		return Orange(text)
	elseif perc < .75 then
		return Yellow(text)
	else
		return Green(text)
	end
end

do
	local header_list = {}
	local skill_frame_shown

	local function SearchForCooldowns()
		addon.Debug("in SearchForCooldowns")

		if not skill_frame_shown then
			return
		end
		------------------------------------------------------------------------------
		-- Save the current state of the TradeSkillFrame so it can be restored after we muck with it.
		------------------------------------------------------------------------------
		local have_materials = _G.TradeSkillFrame.filterTbl.hasMaterials
		local have_skillup = _G.TradeSkillFrame.filterTbl.hasSkillUp

		if have_materials then
			_G.TradeSkillFrame.filterTbl.hasMaterials = false
			_G.TradeSkillOnlyShowMakeable(false)
		end

		if have_skillup then
			_G.TradeSkillFrame.filterTbl.hasSkillUp = false
			_G.TradeSkillOnlyShowSkillUps(false)
		end
		_G.SetTradeSkillInvSlotFilter(0, 1, 1)
		_G.TradeSkillUpdateFilterBar()
		_G.TradeSkillFrame_Update()

		------------------------------------------------------------------------------
		-- Expand headers and scan for cooldowns.
		------------------------------------------------------------------------------
		table.wipe(header_list)

		for skill_index = _G.GetNumTradeSkills(), 1, -1 do
			local entry_name, entry_type, _, is_expanded = _G.GetTradeSkillInfo(skill_index)

			if (entry_type == "header" or entry_type == "subheader") and not is_expanded then
				header_list[entry_name] = true
				_G.ExpandTradeSkillSubClass(skill_index)
			end
		end

		for skill_index = 1, _G.GetNumTradeSkills() do
			local entry_name, entry_type = _G.GetTradeSkillInfo(skill_index)

			if entry_type ~= "header" and entry_type ~= "subheader" then
				local spell_id = tonumber(_G.GetTradeSkillRecipeLink(skill_index):match("^|c%x%x%x%x%x%x%x%x|H%w+:(%d+)"))

				if COOLDOWN_SPELL_MAP[spell_id] then
					local cooldown = _G.GetTradeSkillCooldown(skill_index)
					local cooldown_entry = db.cooldowns[spell_id]

					if not cooldown_entry then
						cooldown_entry = {}
						db.cooldowns[spell_id] = cooldown_entry
					end
					addon.Debug(entry_name .. ", cd is " .. tostring(cooldown))

					local mapped_name = COOLDOWN_SPELL_MAP[spell_id]

					if type(mapped_name) == "string" then
						cooldown_entry.displayName = mapped_name
					else
						cooldown_entry.displayName = entry_name
					end

					if cooldown then
						cooldown_entry.endtime = _G.time() + math.floor(cooldown)
					else
						cooldown_entry.endtime = 0
					end
				else
					-- Just in case we've stored an entry for a spell that was removed from the game.
					db.cooldowns[spell_id] = nil
				end
			end
		end
		------------------------------------------------------------------------------
		-- Restore the state of the things we changed.
		------------------------------------------------------------------------------
		for skill_index = _G.GetNumTradeSkills(), 1, -1 do
			local entry_name, _, _, is_expanded = _G.GetTradeSkillInfo(skill_index)

			if header_list[entry_name] then
				_G.CollapseTradeSkillSubClass(skill_index)
			end
		end
		_G.TradeSkillFrame.filterTbl.hasMaterials = have_materials
		_G.TradeSkillOnlyShowMakeable(have_materials)
		_G.TradeSkillFrame.filterTbl.hasSkillUp = have_skillup
		_G.TradeSkillOnlyShowSkillUps(have_skillup)

		_G.TradeSkillUpdateFilterBar()
		_G.TradeSkillFrame_Update()

	end

	-- I don't actually care about what updated. Just that it did. I only turn on monitoring
	-- for this event when a SPELL_CAST_START has happened for a spell (tradeskill) that has a
	-- cooldown has occurred. And only when it isn't an enchant.
	function BTC:BAG_UPDATE(...)
		addon.Debug("BAG_UPDATE")
		BTC:UnregisterEvent("BAG_UPDATE")

		if skill_frame_shown then
			addon.Debug("SearchForCooldowns")
			SearchForCooldowns()
		elseif not db.options.hideReminders then
			addon.Print(L["CLOSED_TOO_EARLY"])
		end
		addon.updateText()
	end

	function BTC:TRADE_SKILL_SHOW(eventName, ...)
		skill_frame_shown = true
		addon.Debug("In " .. eventName)

		if _G.IsTradeSkillLinked() or _G.IsTradeSkillGuild() then
			addon.Debug("Someone else's tradeskill")
			return
		end
		SearchForCooldowns()
		addon.updateText()
	end

	function BTC:TRADE_SKILL_CLOSE(...)
		skill_frame_shown = false
	end
end -- do-block

function BTC:COMBAT_LOG_EVENT_UNFILTERED(_, _, eventType, _, _, srcName, _, _, _, _, spellId, ...)
	local now = time()
	if ((srcName ~= nil) and
		(srcName == BTC.me) and
		(eventType == "SPELL_CAST_START")) then

		local dbc = db.cooldowns
		-- First, handle the special case items with cooldowns
		--[[ Removed at some point...
		if ( spellId == 19566 ) then
			dbc[spellId].displayName = L["Salt Shaker"]
			dbc[spellId].endtime = now + 252000 ]] --
		if (spellId == 21935) then -- Snowmaster 9000
			dbc[spellId].displayName = L["Snowmaster 9000"]
			dbc[spellId].endtime = now + 72000
		elseif (spellId == 26265) then -- Create Elune Stone
			dbc[spellId].displayName = L["Elune's Lantern"]
			dbc[spellId].endtime = now + 72000
		elseif (COOLDOWN_SPELL_MAP[spellId] ~= nil) then
			if (db.options.hideReminders == false) then
				addon.Print(L["COOLDOWN_DETECTED"])
			end
			-- Right now, all spells with cooldowns place a new item into your bags. Right at the second this
			-- code executes, the player will still be in mid-spell-cast of the spell in question. So in order
			-- to determine when to check for a new cooldown, I'm now going to register for a BAG_UPDATE (the
			-- item hitting my bags), and then scan for new skills.
			BTC:RegisterEvent("BAG_UPDATE")
		else
			addon.Debug("Detected spellcast that wasn't in list. spellid = " .. spellId)
		end
	end
end

function BTC_checkSkills()
	-- local prof1, prof2, arch, fish, cook, firstAid = GetProfessions();
	primary[1], primary[2], secondary[1], secondary[2], secondary[3], secondary[4] = GetProfessions();
	-- local name, texture, rank, maxRank, numSpells, spelloffset, skillLine = GetProfessionInfo(index);
	for i = 1, 2 do
		if (primary[i]) then
			local skillName, _, skillRank, skillMaxRank = GetProfessionInfo(primary[i]);
			db.primary[skillName] = {}
			db.primary[skillName].currlevel = skillRank
			db.primary[skillName].maxlevel = skillMaxRank
		end
	end
	for i = 1, 4 do
		if (secondary[i]) then
			local skillName, _, skillRank, skillMaxRank = GetProfessionInfo(secondary[i]);
			db.secondary[skillName] = {}
			db.secondary[skillName].currlevel = skillRank
			db.secondary[skillName].maxlevel = skillMaxRank
		end
	end
end

local function BTC_MouseHandler(cell, arg, button)
	addon.Debug("Received .. " .. arg)
	-- n~Mazzletov
	-- s~Mazzletov~27008
	local entryType = string.sub(arg, 1, 1)
	local toon, spell, spellName, _
	if (entryType == "c") then
		_, _, _, spellName = string.find(arg, "(%a)~(.+)")
		addon.Debug("spell - >>" .. spellName .. "<<")
		local blort = GetSpellInfo(spellName)
		addon.Debug("blort - " .. blort)
		-- _G.CastSpellByName( "Leatherworking" )
		-- CastSpellByName( blort )
	else
		if (entryType == "n") then
			_, _, _, toon = string.find(arg, "(%a)~(.+)")
			addon.Debug(toon)
		elseif (entryType == "s") then
			_, _, _, toon, spell = string.find(arg, "(%a)~(.-)~(%d+)")
			addon.Debug(toon .. " / " .. spell)
			spell = spell + 0
		end
		if (IsAltKeyDown()) then
			local dbf = BTC.db.factionrealm
			if (button == "LeftButton") then
				addon.Debug("alt-left-click received")
				-- remove player or cooldown from saved values
				if (entryType == "n") then
					for k, v in pairs(dbf[toon].cooldowns) do
						dbf[toon].cooldowns[k] = nil
					end
				elseif (entryType == "s") then
					dbf[toon].cooldowns[spell] = nil
				end
			elseif (button == "RightButton") then
				addon.Debug("alt-right-click received")
				addon.Debug("eT = " .. entryType)
				if (entryType == "n") then
					-- if player, "unhide" all that player's entries
					for k, v in pairs(dbf[toon].cooldowns) do
						v.hide = nil
					end
				elseif (entryType == "s") then
					-- if cooldown, "hide" that cooldown
					dbf[toon].cooldowns[spell].hide = true
					addon.Debug("hiding " .. spell .. " in " .. toon)
				end
			end
		end
	end
	addon.updateText()
	tooltip:Hide()
end

local function DrawTooltip()
	tooltip:Hide()
	tooltip:Clear()

	local dbf = BTC.db.factionrealm
	local lineNum = 0
	local mouseString = ""

	-- Show my primary skills
	tooltip:AddHeader(White(L["Professions :"]))
	for k, v in pairs(db.primary) do
		if (v.currlevel == v.maxlevel) then
			lineNum = tooltip:AddLine(Yellow("   " .. k), " ", Yellow("(" .. v.currlevel .. "/" .. v.maxlevel .. ")"))
		else
			lineNum = tooltip:AddLine(Yellow("   " .. k), " ",
				Yellow("(") .. Red(v.currlevel) .. Yellow("/") .. Green(v.maxlevel) .. Yellow(")"))
		end
		mouseString = "c~" .. k
		-- tooltip:SetLineScript( lineNum, "OnMouseDown", BTC_MouseHandler, mouseString )
	end
	-- Show Secondary
	if (db.options.showSecondary == true) then
		tooltip:AddLine(" ")
		tooltip:AddHeader(White(L["Secondary Skills :"]))
		for k, v in pairs(db.secondary) do
			if (v.currlevel == v.maxlevel) then
				lineNum = tooltip:AddLine(Yellow("   " .. k), " ", Yellow("(" .. v.currlevel .. "/" .. v.maxlevel .. ")"))
			else
				lineNum = tooltip:AddLine(Yellow("   " .. k), " ",
					Yellow("(") .. Red(v.currlevel) .. Yellow("/") .. Green(v.maxlevel) .. Yellow(")"))
			end
			mouseString = "c~" .. k
			-- tooltip:SetLineScript( lineNum, "OnMouseDown", BTC_MouseHandler, mouseString )
		end
	end

	-- cooldowns
	tooltip:AddLine(" ")
	tooltip:AddHeader(White(L["Cooldowns :"]))
	for k, v in pairs(dbf) do
		if (next(v.cooldowns) ~= nil) then
			lineNum = tooltip:AddLine(Yellow("   " .. k))
			mouseString = "n~" .. k
			tooltip:SetLineScript(lineNum, "OnMouseDown", BTC_MouseHandler, mouseString)
		end
		local t = {}
		local Displayed = {}
		local Hidden = {}
		for i, a in pairs(v.cooldowns) do
			local blort = a.displayName
			if (a.hide ~= nil) then
				Hidden[blort] = true
			end
			if not (Displayed[blort]) then
				Displayed[blort] = true
				t[blort] = {}
				local now = time()
				local timeleft = a.starttime
				if ((a.endtime == 0) or (now > a.endtime)) then
					t[blort].outString = Green(L["Ready"])
					-- lineNum = tooltip:AddLine( Yellow( "      " .. a.displayName ), " ", Green( L["Ready"] ) )
				else
					local pretty = BTC_prettyTime(a.endtime - now)
					t[blort].outString = Red(pretty)
					-- lineNum = tooltip:AddLine( Yellow( "      " .. a.displayName ), " ", Red( pretty ) )
				end
				t[blort].mouseString = "s~" .. k .. "~" .. i
				-- tooltip:SetLineScript( lineNum, "OnMouseDown", BTC_MouseHandler, mouseString )
			end
		end
		for i, a in pairs(t) do
			if not (Hidden[i]) then
				lineNum = tooltip:AddLine(Yellow("      " .. i), " ", a.outString)
				tooltip:SetLineScript(lineNum, "OnMouseDown", BTC_MouseHandler, a.mouseString)
			end
		end
	end
	-- Footer
	tooltip:AddLine(" ")
	lineNum = tooltip:AddLine("b")
	tooltip:SetCell(lineNum, 1, L["Click for Configuration and Help"], "LEFT", 3)
end

local LDB = LibStub("LibDataBroker-1.1", true)
local launcher = LDB:NewDataObject("BTC", {
	type = "data source",
	text = "",
	label = "Broker_TradeCooldowns",
	icon = "Interface\\Addons\\Broker_TradeCooldowns\\icon",
	OnClick = function(clickedframe, button)
		BTC:ShowConfig()
		addon.updateText()
	end,
	OnEnter = function(self)
		tooltip = LQT:Acquire(ADDON_NAME_SHORT .. "TT", 3, "LEFT", "CENTER", "RIGHT") -- or GameTooltip
		tooltip:SetScale(db.options.tipscale)
		DrawTooltip()
		tooltip:SetAutoHideDelay(.1, self)
		tooltip:EnableMouse()
		tooltip:SmartAnchorTo(self)
		tooltip:Show()
	end,
	OnLeave = function()
	-- this should get handled by the autohide delay stuff now
	end,
})

-- options stuff
local displayOptions = nil
local function giveDisplay()
	if (not displayOptions) then
		displayOptions = {
			order = 4,
			type = "group",
			name = L["Display Options"],
			desc = L["DISPLAY_OPTIONS_DESC"],
			args = {
				header1 = {
					order = 41,
					type = "header",
					name = L["Display Options"],
				},
				desc = {
					order = 42,
					type = "description",
					name = L["DISPLAY_OPTIONS_DESC"],
				},
				tooltipscale = {
					order = 43,
					type = "range",
					name = L["Tooltip Scale"],
					desc = L["TOOLTIP_SCALE_DESC"],
					min = .7,
					max = 1.3,
					step = .05,
					bigStep = .05,
					get = function() return db.options.tipscale end,
					set = function(info, v)
						db.options.tipscale = v
						tooltip:SetScale(v)
					end,
				},
				debug = {
					order = 44,
					type = "toggle",
					name = L["Debug"],
					desc = L["DEBUG_DESC"],
					get = function() return db.options.debug end,
					set = function() db.options.debug = not db.options.debug end,
				},
				showsecondary = {
					order = 46,
					type = "toggle",
					name = L["Show Secondary"],
					desc = L["SHOW_SECONDARY_DESC"],
					get = function() return db.options.showSecondary end,
					set = function() db.options.showSecondary = not db.options.showSecondary end,
				},
				hidereminders = {
					order = 47,
					type = "toggle",
					name = L["Hide Reminders"],
					desc = L["HIDE_REMINDERS_DESC"],
					get = function() return db.options.hideReminders end,
					set = function() db.options.hideReminders = not db.options.hideReminders end,
				},
				header2 = {
					order = 200,
					type = "header",
					name = L["Minimap Icon Options"],
				},
				desc2 = {
					order = 201,
					type = "description",
					name = L["MINIMAP_OPTIONS_DESC"],
				},
				minimap_icon = {
					order = 202,
					type = "toggle",
					width = "full",
					name = L["Show Minimap Icon"],
					desc = L["Draws the icon on the minimap."],
					get = function()
						return not db.options.minimap_icon.hide
					end,
					set = function(info, value)
						db.options.minimap_icon.hide = not value
						LDBIcon[value and "Show" or "Hide"](LDBIcon, ADDON_NAME_SHORT)
					end,
				},
				header3 = {
					order = 301,
					type = "header",
					name = L["Help and Usage Notes"],
				},
				help1 = {
					order = 302,
					type = "description",
					name = "- " .. L["HELP_STRING_0"],
				},
				help2 = {
					order = 303,
					type = "description",
					name = "- " .. L["HELP_STRING_1"],
				},
				help3 = {
					order = 304,
					type = "description",
					name = "- " .. L["HELP_STRING_2"],
				},
				help4 = {
					order = 305,
					type = "description",
					name = "- " .. L["HELP_STRING_3"],
				},
				help5 = {
					order = 306,
					type = "description",
					name = "- " .. L["HELP_STRING_4"],
				},
			},
		}
	end
	return displayOptions
end

local textOptions = nil
local function giveText()
	if (not textOptions) then
		textOptions = {
			order = 4,
			type = "group",
			name = L["Text Options"],
			desc = L["TEXT_OPTIONS_DESC"],
			args = {
				header1 = {
					order = 41,
					type = "header",
					name = L["Text Options"],
				},
				desc = {
					order = 42,
					type = "description",
					name = L["TEXT_OPTIONS_DESC"],
				},
				showtext = {
					order = 50,
					type = "toggle",
					name = L["Show Text"],
					desc = L["SHOW_TEXT_DESC"],
					get = function() return db.options.showText end,
					set = function()
						db.options.showText = not db.options.showText
						addon.updateText()
					end,
				},
				showcounters = {
					order = 51,
					type = "toggle",
					name = L["Show Counters"],
					desc = L["SHOW_COUNTERS_DESC"],
					get = function() return db.options.showCounters end,
					set = function()
						db.options.showCounters = not db.options.showCounters
						addon.updateText()
					end,
					disabled = function()
						if (db.options.showText) then
							return false
						else
							return true
						end
					end,
				},
				textcooldown = {
					order = 52,
					type = "select",
					name = L["Select Cooldown to Display"],
					desc = L["SHOW_TEXTCD_DESC"],
					get = function() return db.options.textCooldown end,
					set = function(info, value)
						db.options.textCooldown = value
						addon.updateText()
					end,
					values = function()
						local dbf = BTC.db.factionrealm
						local t = {}
						local index = ""
						for k, v in pairs(dbf) do
							local Displayed = {}
							for i, j in pairs(v.cooldowns) do
								if (Displayed[i] ~= false) then
									if (j.hide ~= nil) then
										Displayed[i] = false
									else
										if (Displayed[i] == true) then
											-- already displayed, don't do anything
										else
											index = k .. ":" .. i
											t[index] = k .. ":" .. j.displayName
											Displayed[i] = true
										end
									end
								end
							end
						end
						return t
					end,
					disabled = function()
						if (db.options.showText) then
							if (db.options.showCounters) then
								return true
							else
								return false
								--[[
								if ( next( db.cooldowns ) ~= nil ) then
									return false
								else
									return true
								end ]] --
							end
						else
							return true
						end
					end,
				},
			},
		}
	end
	return textOptions
end

local modularOptions = {}
local options = nil
local function fullOptions()
	if (not options) then
		options = {
			type = "group",
			name = ADDON_NAME_SHORT,
			args = {
				general = {
					order = 1,
					type = "group",
					name = L["General Information"],
					desc = L["GENERAL_INFO_DESC"],
					args = {
						header1 = {
							order = 11,
							type = "header",
							name = L["General Information"],
						},
						desc = {
							order = 12,
							type = "description",
							name = L["GENERAL_INFO_DESC"],
						},
						author = {
							order = 13,
							type = "description",
							name = L["Author : "] .. addonauthor .. "\n",
						},
						version = {
							order = 14,
							type = "description",
							name = L["Version : "] .. addonversion .. "\n",
						},
						builddate = {
							order = 15,
							type = "description",
							name = L["Build Date : "] .. builddate .. "\n",
						},
						dbversion = {
							order = 15,
							type = "description",
							name = L["Database Version : "] .. db.options.dbVersion .. "\n",
						},
						translators = {
							order = 16,
							type = "description",
							name = string.format(L["Helpful Translators (thank you) : %s"],
								"whynona2(frFR), SunTsu(deDE), Dessa(deDE), " ..
									"amunra(koKR), xska(ruRU), <CWDG>tnt2ray(zhCN), " ..
									"NightOw1(zhTW), andy52005(zhTW)\n")
						},
					},
				},
			},
		}
	end

	for k, v in pairs(modularOptions) do
		options.args[k] = (type(v) == "function") and v() or v
	end

	return options
end

local AceConfig = LibStub("AceConfig-3.0")
local AceConfigReg = LibStub("AceConfigRegistry-3.0")
local AceConfigDialog = LibStub("AceConfigDialog-3.0")

function addon:SetupOptions()
	AceConfigReg:RegisterOptionsTable(ADDON_NAME_SHORT, fullOptions)
	self.optionsFrame = AceConfigDialog:AddToBlizOptions(ADDON_NAME_SHORT, nil, nil, "general")

	-- Fill up our modular options...
	self:RegisterModuleOptions("Display", giveDisplay(), L["Display Options"])
	self:RegisterModuleOptions("Text", giveText(), L["Text Options"])
end

function addon:RegisterModuleOptions(name, optionsTable, displayName)
	modularOptions[name] = optionsTable
	self.optionsFrame[name] = AceConfigDialog:AddToBlizOptions(ADDON_NAME_SHORT, displayName, ADDON_NAME_SHORT, name)
end

function BTC:ShowConfig()
	InterfaceOptionsFrame_OpenToCategory(self.optionsFrame["Display"])
end

function BTC.updateText()
	addon.Debug("In updateText")
	local newStr = " "
	if (db.options.showText) then
		if (db.options.showCounters) then
			-- show counter, like (8/10). Color text based on percent of CDs available
			local numCDs = 0
			local numAvail = 0
			local dbf = BTC.db.factionrealm
			for k, v in pairs(dbf) do
				local t = {}
				local Displayed = {}
				local Hidden = {}
				for i, a in pairs(v.cooldowns) do
					local blort = a.displayName
					if (a.hide ~= nil) then
						Hidden[blort] = true
					end
					if not (Displayed[blort]) then
						Displayed[blort] = true
						local now = time()
						local timeleft = a.starttime
						if ((a.endtime == 0) or (now > a.endtime)) then
							t[blort] = true
						else
							t[blort] = false
						end
					end
				end
				for i, a in pairs(t) do
					if not (Hidden[i]) then
						numCDs = numCDs + 1
						if (a == true) then
							numAvail = numAvail + 1
						end
					end
				end
			end
			newStr = "(" .. numAvail .. "/" .. numCDs .. ")"
			newStr = PercentColorText(numAvail, numCDs, newStr)
			addon.Debug("Counters = " .. newStr)
		else
			-- Show an individual cooldown, not a counter
			if (db.options.textCooldown ~= 0) then
				local dbf = addon.db.factionrealm
				-- Should be in the form: toonName:CooldownSpellID
				addon.Debug("textCooldown = " .. db.options.textCooldown)
				local selToon, selCD = (":"):split(db.options.textCooldown)
				if ((selToon == nil) or (selToon == "") or (selCD == nil)) then
					-- for some reason, its possible to have a blank toon here. No idea why. THrow an error.
					addon.Print(L["Couldn't assign value. Please file a bug ticket."])
					addon.Print(L["Include the following in the ticket:"])
					if (selToon == nil) then addon.Print("selToon is nil") end
					if (selToon == "") then addon.Print("selToon is blank") end
					if (selCD == nil) then addon.Print("selCD is nil") end
				else
					addon.Debug("toon = " .. selToon .. ", cd = " .. selCD)
					for k, v in pairs(dbf) do
						addon.Debug("k=" .. k)
						if (k == selToon) then
							addon.Debug("matched toon")
							for i, j in pairs(v.cooldowns) do
								addon.Debug("i=" .. i)
								selCD = selCD + 0
								if (i == selCD) then
									addon.Debug("matched spell")
									newStr = Yellow(j.displayName) .. ":"
									local now = time()
									if ((j.endtime == 0) or (now > j.endtime)) then
										newStr = newStr .. Green(L["Ready"])
									else
										local pretty = BTC_prettyTime(j.endtime - now)
										newStr = newStr .. Red(pretty)
									end
								end
							end
						end
					end
				end
			end
		end
	end
	launcher.text = newStr
end

function BTC_upgradeDB()
	local olddb = BTC.db.factionrealm[BTC.me]
	-- This checks the current DB version the person has, and upgrades as necessary
	if not (olddb.options) then
		-- They have the original version
		olddb.options = {}
		olddb.options = {
			dbVersion = "1.0",
			debug = false,
			showWeapons = false,
			showSecondary = false,
			showClass = false,
			showArmor = false,
			showLanguages = false,
		}
		olddb.weapons = {}
		olddb.primary = {}
		olddb.secondary = {}
		olddb.class = {}
		olddb.languages = {}
		olddb.armor = {}
		olddb.skills = nil
		addon.Print(string.format(L["Database upgraded to %s"], "1.0"))
	end
	if (olddb.options.dbVersion == "1.0") then
		olddb.options.tipscale = 1
		olddb.options.dbVersion = "1.1"
		addon.Print(string.format(L["Database upgraded to %s"], "1.1"))
	end
	if (olddb.options.dbVersion == "1.1") then
		olddb.options.showText = false
		olddb.options.textCooldown = 0
		olddb.options.dbVersion = "1.2"
		addon.Print(string.format(L["Database upgraded to %s"], "1.2"))
	end
	if (olddb.options.dbVersion == "1.2") then
		olddb.options.showClass = nil
		olddb.options.showArmor = nil
		olddb.options.showLanguages = nil
		olddb.languages = nil
		olddb.class = nil
		olddb.armor = nil
		olddb.options.dbVersion = "1.3"
		addon.Print(string.format(L["Database upgraded to %s"], "1.2"))
	end
	-- pretty much want to do this no matter what...
	local OldCleared = false
	for k, v in pairs(olddb.cooldowns) do
		if (COOLDOWN_SPELL_MAP[k] == nil) then
			olddb.cooldowns[k] = nil
			OldCleared = true
		end
	end
	if (OldCleared == true) then
		addon.Print(L["Cleared out old cooldowns."])
	end
	-------------------------------------------------
	if ((olddb.options.dbVersion == "1.3") or (olddb.options.dbVersion == "1.4")) then
		olddb.options.hideReminders = false
		olddb.options.dbVersion = "1.5"
		addon.Print(string.format(L["Database upgraded to %s"], "1.5"))
	end
	if (olddb.options.dbVersion == "1.5") then
		olddb.weapons = nil
		olddb.secondary[L["Riding"]] = nil
		olddb.options.showWeapons = nil
		olddb.options.dbVersion = "1.6"
		addon.Print(string.format(L["Database upgraded to %s"], "1.6"))
	end
	if (olddb.options.dbVersion == "1.6") then
		olddb.options.minimap_icon = nil
		olddb.options.minimap_icon = {}
		olddb.options.minimap_icon.hide = true
		olddb.options.dbVersion = "1.7"
		addon.Print(string.format(L["Database upgraded to %s"], "1.7"))
	end
	if (olddb.options.dbVersion == "1.7") then
		olddb.options.showCounters = false
		olddb.options.dbVersion = "1.8"
		addon.Print(string.format(L["Database upgraded to %s"], "1.8"))
	end
	if (olddb.options.dbVersion == "1.8") then
		for k, v in pairs(olddb.cooldowns) do
			olddb.cooldowns[k] = nil
		end
		olddb.cooldowns = nil
		olddb.cooldowns = {}
		olddb.options.dbVersion = "1.9"
		addon.Print(string.format(L["Database upgraded to %s"], "1.9"))
		addon.Print(L["This version upgrade required a clear of cooldowns"])
		addon.Print(L["to remove possible corruption. Please rescan your professions."])
	end
end

function addon.stopScanning()
	if (BTC_ScanFrame) then
		BTC_ScanFrame:SetScript("OnUpdate", nil);
	end
end

function addon.scanSkills()
	-- on first load, there's no real good way to check if the player's skills
	-- are available for scanning yet. So just scan every 5 seconds until they are
	if (not BTC_ScanFrame) then
		BTC_ScanFrame = CreateFrame("Frame")
	end

	local totalElapsed = 0
	local checkInterval = 60
	local gotFirst = false

	BTC_ScanFrame:SetScript("OnUpdate", function(self, elapsed)
		totalElapsed = totalElapsed + elapsed
		if (totalElapsed < checkInterval) then
			return;
		end
		totalElapsed = 0;
		BTC_checkSkills()
		addon.updateText()
		return
	end)
end

-- on init
function BTC:OnInitialize()
end

function BTC:OnEnable()
	-- Called when the addon is loaded
	self.db = LibStub("AceDB-3.0"):New("BTC_DB", nil, "Default")
	if (not self.db.factionrealm) then
		self.db.factionrealm = {}
	end

	BTC.me = UnitName("player")

	-- do we have any data yet for this toon? If not, init its table
	if not (self.db.factionrealm[BTC.me]) then
		self.db.factionrealm[BTC.me] = {
			primary = {},
			secondary = {},
			class = {},
			languages = {},
			armor = {},
			cooldowns = {},
			options = {
				dbVersion = "1.8",
				debug = false,
				showSecondary = false,
				showText = false,
				showCounters = false,
				hideReminders = false,
				textCooldown = 0,
				tipscale = 1,
				minimap_icon = {
					hide = true,
				},
			}
		}
	else
		-- check if we need to upgrade the db
		BTC_upgradeDB()
	end

	db = self.db.factionrealm[BTC.me]

	addon.Debug("In OnEnable")

	BTC:SetupOptions()

	-- Check current weapons/languages
	addon.scanSkills()
	addon.updateText()

	-- Event listeners
	BTC:RegisterEvent("TRADE_SKILL_SHOW")
	BTC:RegisterEvent("TRADE_SKILL_CLOSE")
	BTC:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	BTC:RegisterEvent("CHAT_MSG_SKILL", BTC_checkSkills)

	-- libdbicon stuff
	if LDBIcon then
		LDBIcon:Register(ADDON_NAME_SHORT, launcher, db.options.minimap_icon)
	end
end
